/* Copyright (C) 1998 Oregon Graduate Institute of Science & Technology
 * Portions copyright (c) 1995 Brown University.
 * Copyright (c) 1995 The Regents of the University of California.
 * 
 * See the file COPYRIGHT, which should have been supplied with this
 * package, for licensing details.  We may be contacted through email
 * at <quasar-help@cse.ogi.edu>.
 */

/*
 * main.c --
 *
 *      Main procedure
 *
 */

#include "global.h"
#include "video.h"
#include "util.h"
#include "gdith.h"
#include "stream_play.h"

#include <string.h>
#include <signal.h>
#include <errno.h>
#include <syslog.h>
#include <dbug.h>
#include <pipeline/channel.h>


#ifdef HAVE_LAUNCHER
#include <pipeline/init_modules.h>
#endif /* HAVE_LAUNCHER */

#include <getopt.h>

/* Define buffer length. */
#define BUF_LENGTH 4096



static void
usage()
{
        fprintf(stderr, 
                "Usage:    mpegdecode input output\n"
                "     or   mpegdecode -i input      if mpegdecode is used before  \"\\|\"\n"
                "     or   mpegdecode -o output     if mpegdecode is used after   \"\\|\"\n"
                "     or   mpegdecode               if mpegdecode is used between \"\\|\"\n"
		);
        exit(1);
}


/*
 *--------------------------------------------------------------
 *
 * mpegdecode
 *
 *        initialize the decoder thread, adapted from main().
 *
 *--------------------------------------------------------------
 */

#ifdef HAVE_CHANNELS

#ifdef HAVE_LAUNCHER
void
mpegdecode(int argc, char **argv) {

  channel_src *channelin = NULL;
  channel_dst *channelout = NULL;

  DataStream *theSource = NULL;
  VidStream *theStream = NULL;
  dbuf *dbuf;
  char *inputName = "-";
  int ditherType;

  int opt = 0;

  DBUG_ENTER();

  // module_init takes care of -iqueue and -oqueue flags if the launcher has
  // added any. Thus it is possible that packin or packout gets connected 
  // to a channel in the function module_init.
 argc = module_init(argc, argv, &channelin, ARG_chunk, &channelout, ARG_chunk);
 // when module_init returns, argc and argv have been modified if any -iqueue
 // or -oqueue have been found. The module needs now to parse its argument and
 // connect packin or (and) packout to channels if they are not already.

 // getopt_long doesn't support multi-threading, thus we need to use a mutex.
  pthread_mutex_lock(&launcher_mutex);

  while (1) {
    int option_index = 0;
    struct option long_options[] =
    {
      {"help", 0, 0, 0},
      {0, 0, 0, 0}
    };
    int c = getopt_long(argc, argv, "i:o:",long_options, &option_index);
    if (c == -1)
      break;
    switch (c) {
    case 0:
      if (option_index == 0) usage();
    case 'i':
      channelin = open_channel_src(optarg, ARG_chunk);
      break;
    case 'o':
      channelout = open_channel_dst(optarg, ARG_chunk);
      break;
    default:
      usage();
    }
  }
  
 opt =  optind;
 optind = 0;  // to "reset" getopt!!!
 
 // we are done with getopt so we can release the mutex.
 pthread_mutex_unlock(&launcher_mutex);


  
 if ((argc - opt) == 2) 
	 {
	  channelout = open_channel_dst(argv[opt + 1], ARG_chunk);
	  argc--;
	 }
  if ((argc - opt) == 1) 
	  channelin = open_channel_src(argv[opt], ARG_chunk);

  
  if ((channelin == NULL) || (channelout == NULL))
    usage();


#else 


void
mpegdecode(channel_src *channelin, channel_dst *channelout)
{
  DataStream *theSource = NULL;
  VidStream *theStream = NULL;
  dbuf *dbuf;
  char *inputName = "-";
  int ditherType;
  DBUG_ENTER();

#endif /* HAVE_LAUNCHER */

  DBUG_KEYWORDS("display",NULL);
  /* Set up tables. */
  InitEverything();
  /* Create data stream. */
  theSource = NewDataStream((unsigned int) BUF_LENGTH);
  theSource->channelin = channelin;
  theSource->channelout = channelout;
  theSource->input = NULL;
  ditherType = InitDithering();
  theStream = NewVidStream(theSource, inputName, ditherType);
  theStream->rate_ticks = -1;	/* Activate frame rate control. */
  dbuf = new_raw_dbuf(sizeof(VidStream *));
  dbGROW(dbuf,sizeof(VidStream *));
  *(VidStream **)dbCUR(dbuf) = theStream;
  chSEND_free(channelout,dbuf);

  if (!mpegVidRsrc(theStream)) {
    fprintf(stderr, "Skipping movie \"%s\" - not an MPEG stream\n", inputName);
    goto finishUp;
  }
  while (!theStream->film_has_ended) {
    mpegVidRsrc(theStream);
  }
finishUp:
  chSEND_free(channelout,new_raw_dbuf(0));
  while(theStream->film_has_ended) {
    sleep(1);
  }
  
  DestroyVidStream(theStream);
  DestroyDataStream(theSource);

  DBUG_VOID_RETURN();
}
 
#endif // HAVE_CHANNELS
